"""
Simulation module for BEAMZ - Contains FDTD simulation and field operations.
"""

from beamz.design.meshing import RegularGrid
from beamz.simulation.core import Simulation

__all__ = ['RegularGrid', 'Simulation'] 