LIGHT_SPEED = 299792458.0 # m/s
VAC_PERMITTIVITY = 8.85e-12  # Vacuum permittivity
VAC_PERMEABILITY = 1.256e-6  # Vacuum permeability
EPS_0 = 8.85e-12
MU_0 = 1.256e-6

# SI units
um = 1e-6
µm = 1e-6
nm = 1e-9
